'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from dependency_injector.wiring import inject, Provide
from altspell_plugins import PluginBase
from ..services import SpellingSystemService
from ..containers import Container


@inject
def populate_spelling_system_table(
    plugin_instance: PluginBase,
    spelling_system_service: SpellingSystemService = Provide[Container.spelling_system_service]
):
    """Populate spelling system table with spelling system"""
    spelling_system_service.add_spelling_system(
        plugin_instance.name,
        plugin_instance.version,
        plugin_instance.pretty_name,
        plugin_instance.facts
    )



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2024  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from functools import wraps
import requests
from flask import current_app, request, jsonify


def _verify_hcaptcha(token):
    secret_key = current_app.config['HCAPTCHA_SECRET_KEY']
    verification_url = 'https://api.hcaptcha.com/siteverify'

    # prepare payload for POST request
    payload = {
        'secret': secret_key,
        'response': token
    }

    resp = requests.post(verification_url, data=payload, timeout=10)

    return resp.json().get('success') is True

def require_hcaptcha(func):
    """
    A decorator applied to routes that require hCaptcha validation.

    The decorator checks the JSON request object for a valid hCaptcha token before allowing the
    request to proceed. The hCaptcha token is expected in the JSON key: 'hcaptcha_token'. hCaptcha
    verification can be bypassed by setting ENABLE_HCAPTCHA to False in the app instance config.

    Args:
        func: Function for which hCaptcha is required.

    Returns:
        function: If the hCaptcha is verified successfully, func is returned.
        Response: If the hCaptcha is missing or fails to verify, a '400 Bad Request' Response is
                  returned.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if current_app.config['ENABLE_HCAPTCHA'] is not True:
            return func(*args, **kwargs)

        data = request.json
        hcaptcha_token = data.get('hcaptcha_token')

        if not hcaptcha_token:
            return jsonify({"error": "hCaptcha token is missing"}), 400

        # verify hCaptcha token
        if _verify_hcaptcha(hcaptcha_token) is False:
            return jsonify({"error": "hCaptcha verification failed"}), 400

        # if valid hCaptcha, proceed with the request
        return func(*args, **kwargs)

    return wrapper



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2024  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import uuid
import datetime
from typing import List, Optional
from sqlalchemy import Boolean, DateTime, ForeignKey, UniqueConstraint, Integer, String, Uuid, JSON
from sqlalchemy.orm import Mapped, mapped_column, relationship, DeclarativeBase
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles


class Base(DeclarativeBase):  # pylint: disable=missing-class-docstring,too-few-public-methods
    pass

class UTCnow(expression.FunctionElement):
    """A construct representing the current UTC timestamp."""
    type = DateTime()
    inherit_cache = True

@compiles(UTCnow, "postgresql")
def pg_utcnow(_element, _compiler, **_kw):
    """Compiles the `UTCnow` expression to the PostgreSQL-specific SQL syntax for retrieving the
    UTC timestamp."""
    return "date_trunc('second', TIMEZONE('utc', CURRENT_TIMESTAMP))"

@compiles(UTCnow, "mssql")
def ms_utcnow(_element, _compiler, **_kw):
    """Compiles the `UTCnow` expression to the Microsoft SQL-specific SQL syntax for retrieving the
    UTC timestamp."""
    return "GETUTCDATE() AS SMALLDATETIME"

@compiles(UTCnow, "sqlite")
def sqlite_utcnow(_element, _compiler, **_kw):
    """Compiles the `UTCnow` expression to the Sqlite-specific SQL syntax for retrieving the UTC
    timestamp."""
    return "(STRFTIME('%Y-%m-%d %H:%M:%S', 'NOW'))"

class SpellingSystem(Base):  # pylint: disable=too-few-public-methods
    """A table containing the enabled alternate spellings of English."""
    __tablename__ = "spelling_system"
    __table_args__ = (
        UniqueConstraint('name', 'version'),
    )

    id: Mapped[int] = mapped_column(
        Integer,
        primary_key=True,
        doc='Sequence number representing alternative spelling of English.'
    )
    name: Mapped[str] = mapped_column(
        String,
        doc='Name of alternative spelling of English.'
    )
    version: Mapped[str] = mapped_column(
        String,
        doc='Version number of alternative spelling of English'
    )
    pretty_name: Mapped[str] = mapped_column(
        String,
        doc='Pretty name of alternative spelling of English.'
    )
    facts: Mapped[Optional[JSON]] = mapped_column(
        JSON,
        doc='Optional JSON document with miscellaneous info about alternative spelling of English.'
    )

    translations: Mapped[List["Translation"]] = relationship(
        back_populates="spelling_system",
        doc='All translations that use the alternative spelling of English.'
    )

class Translation(Base):  # pylint: disable=too-few-public-methods
    """A table containing the saved translations."""
    __tablename__ = "translation"

    id: Mapped[uuid] = mapped_column(
        Uuid,
        primary_key=True,
        doc='Sequence number representing translation.'
    )
    creation_date: Mapped[datetime.datetime] = mapped_column(
        DateTime(),
        server_default=UTCnow(),
        doc='DateTime representing when the translation was inserted into the database.'
    )
    forward: Mapped[bool] = mapped_column(
        Boolean,
        doc='Boolean representing which direction the translation occurred in. I.e: Either ' \
            'traditional English spelling => alternative English spelling or alternative ' \
            'English spelling => traditional English spelling.'
    )
    traditional_text: Mapped[str] = mapped_column(
        String,
        doc='Text in traditional English spelling.'
    )
    respelled_text: Mapped[str] = mapped_column(
        String,
        doc='Text in alternative English spelling.'
    )
    spelling_system_id: Mapped[int] = mapped_column(
        ForeignKey('spelling_system.id'),
        doc='Sequence number representing alternative spelling of English.'
    )

    spelling_system: Mapped["SpellingSystem"] = relationship(
        back_populates="translations",
        doc='The alternative spelling of English that was used by the translation.'
    )



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2024-2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import importlib
import pkgutil
import os
from importlib.metadata import version
from flask import Flask
from flask_cors import CORS
from sqlalchemy.exc import IntegrityError
from altspell_plugins import PluginBase
from .model import Base


AVAILABLE_PLUGINS = {
    name.removeprefix('altspell_'): importlib.import_module(name)
    for finder, name, ispkg
    in pkgutil.iter_modules()
    if name.startswith('altspell_') and name != 'altspell_plugins' # ignore plugins interface
}

def create_app(test_config=None):
    """Create and configure an instance of the Flask application."""
    app = Flask(__name__, instance_relative_config=True)

    app.config.from_mapping(
        # a default secret that should be overridden by instance config
        SECRET_KEY="dev",
        # store the database in the app instance path
        SQLALCHEMY_DATABASE_URI='sqlite:///' + os.path.join(app.instance_path, 'altspell.db'),
        # maximum number of characters accepted for translation
        TRANSLATION_LENGTH_LIMIT = 20000,
        # enable all plugins by default
        ENABLED_PLUGINS = AVAILABLE_PLUGINS.keys(),
        # disable CAPTCHA for test purposes
        ENABLE_HCAPTCHA = False
    )

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.update(test_config)

    with app.app_context():
        from .containers import Container  # pylint: disable=import-outside-toplevel

        # create container for dependency injection
        container = Container()

    app.container = container

    # configure the database
    db = container.db(model_class=Base)
    db.init_app(app)

    # create the cache
    cache = container.cache()
    cache.init_app(app)

    # create the migration
    migrate = container.migrate()
    migrate.init_app(app, db)

    # allow CORS for all domains on all routes
    CORS(app)

    # ensure the instance folder exists
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    # create the database
    with app.app_context():
        db.create_all()

    app.plugin_instances = {}

    for plugin in app.config['ENABLED_PLUGINS']:
        if plugin in AVAILABLE_PLUGINS:
            plugin_mod = AVAILABLE_PLUGINS[plugin]

            # validate plugin implementation
            if not hasattr(plugin_mod, 'Plugin'):
                app.logger.error(
                    'Enabled plugin excluded for missing "Plugin" attribute: %s', plugin
                )
                continue

            if not issubclass(plugin_mod.Plugin, PluginBase):
                app.logger.error(
                    'Enabled plugin excluded for not being a subclass of PluginBase: %s', plugin
                )
                continue

            # initialize plugin
            app.logger.info('Initializing plugin: %s...', plugin)
            plugin_instance = plugin_mod.Plugin()
            plugin_instance.version = version(plugin_mod.__name__)
            app.plugin_instances[plugin_instance.name] = plugin_instance

            with app.app_context():
                from .utils.populate_spelling_system_table import populate_spelling_system_table  # pylint: disable=import-outside-toplevel
                populate_spelling_system_table(plugin_instance)
        else:
            app.logger.warning('Enabled plugin is not available: %s', plugin)

    # apply the blueprints to the app
    from .blueprints import translation  # pylint: disable=import-outside-toplevel
    app.register_blueprint(translation.bp)

    from .blueprints import spelling_system  # pylint: disable=import-outside-toplevel
    app.register_blueprint(spelling_system.bp)

    return app



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import uuid
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload
from .model import SpellingSystem, Translation
from .exceptions import TranslationNotFoundError, SpellingSystemNotFoundError


class TranslationRepository:
    """Repository for database operations related to translations."""
    def __init__(self, db: SQLAlchemy) -> None:
        self.db = db

    def add(
        self,
        forward: bool,
        traditional_text: str,
        respelled_text: str,
        spelling_system_id: int
    ) -> Translation:
        """
        Add a translation to the database.

        Args:
            forward (bool): True if translated to the alternative spelling system. False if \
                translated to traditional English spelling.
            traditional_text (str): Text in traditional English spelling.
            respelled_text (str): Text in the alternative English spelling system.
            spelling_system_id (int): Id of the alternative spelling system.

        Returns:
            Translation: The translation object added to the database.
        """
        translation = Translation(
            id=uuid.uuid4(),
            forward=forward,
            traditional_text=traditional_text,
            respelled_text=respelled_text,
            spelling_system_id=spelling_system_id
        )
        self.db.session.add(translation)
        try:
            self.db.session.commit()
        except IntegrityError:
            self.db.session.rollback()
        translation = (
            self.db.session.query(Translation)
            .options(selectinload(Translation.spelling_system))
            .filter(Translation.id == translation.id)
            .first()
        )
        return translation

    def get_by_id(self, translation_id: uuid) -> Translation:
        """
        Retrieve a translation by id.

        Args:
            translation_id (uuid): Id of the requested translation.

        Returns:
            Translation: The translation object corresponding to translation_id.
        """
        translation = (
            self.db.session.query(Translation)
            .options(selectinload(Translation.spelling_system))
            .filter(Translation.id == translation_id)
            .first()
        )
        if not translation:
            raise TranslationNotFoundError(translation_id)
        return translation

class SpellingSystemRepository:
    """Repository for database operations related to alternative spelling systems."""
    def __init__(self, db: SQLAlchemy) -> None:
        self.db = db

    def add(self, name: str, version: str, pretty_name: str, facts: str | None) -> SpellingSystem:
        """
        Add an alternative spelling system.

        Args:
            name (str): Name of the alternative spelling system.
            version (str): Version of the alternative spelling system.
            pretty_name (str): Pretty name of the alternative spelling system.
            facts (str): JSON string for miscellaneous structured info about the spelling system
                plugin.

        Returns:
            SpellingSystem: The alternative spelling system object added to the database.
        """

        spelling_system = SpellingSystem(
            name=name,
            version=version,
            pretty_name=pretty_name,
            facts=facts
        )
        self.db.session.add(spelling_system)
        try:
            self.db.session.commit()
        except IntegrityError:
            self.db.session.rollback()
        spelling_system = (
            self.db.session.query(SpellingSystem)
            .filter(SpellingSystem.name == name, SpellingSystem.version == version)
            .first()
        )
        return spelling_system

    def get_all(self):
        """Retrieve a list of enabled alternative spelling systems."""
        return self.db.session.query(SpellingSystem).all()

    def get(self, name: str, version: str) -> SpellingSystem:
        """
        Retrieve an alternative spelling system object by alternative spelling system name and
        version.

        Args:
            name (str): Name of the alternative spelling system.
            version (str): Version of the alternative spelling system.

        Returns:
            SpellingSystem: The alternative spelling system object corresponding to \
                name and version.
        """
        spelling_system = (
            self.db.session.query(SpellingSystem)
            .filter(SpellingSystem.name == name, SpellingSystem.version == version)
            .first()
        )
        if not spelling_system:
            raise SpellingSystemNotFoundError(f"{name} v{version}")
        return spelling_system



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from typing import Type


class NotFoundError(Exception):
    """Exception for database records that cannot be found."""
    entity_name: str

    def __init__(self, entity_id):
        super().__init__(f"{self.entity_name} not found, id: {entity_id}")

class TranslationNotFoundError(NotFoundError):
    """Exception for Translations that cannot be found."""
    entity_name: str = "Translation"

class SpellingSystemNotFoundError(NotFoundError):
    """Exception for spelling systems that cannot be found."""
    entity_name: str = "Spelling system"

class MissingKeyError(Exception):
    """Exception for missing required JSON request body keys."""
    def __init__(self, key_name: str):
        super().__init__(f"Missing key: {key_name}")

class InvalidTypeError(Exception):
    """Exception for JSON request body keys with incorrect types."""
    def __init__(self, key_name: str, cls: Type):
        super().__init__(f"Key '{key_name}' must be of type '{cls.__name__}'")

class EmptyTranslationError(Exception):
    """Exception for empty translation text."""
    def __init__(self):
        super().__init__("Cannot save an empty translation")

class FwdTranslationNotImplementedError(NotImplementedError):
    """Exception for attempted unimplemented forward translations."""
    def __init__(self):
        super().__init__("Forward translation is not implemented for this spelling system")

class BwdTranslationNotImplementedError(NotImplementedError):
    """Exception for attempted unimplemented backward translations."""
    def __init__(self):
        super().__init__("Backward translation is not implemented for this spelling system")

class SpellingSystemUnavailableError(Exception):
    """Exception for attempted translations with unavailable spelling systems."""
    def __init__(self, spelling_system: str):
        super().__init__(f"Spelling system '{spelling_system}' is unavailable")



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2024-2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import uuid
from flask import Blueprint, request
from flask_caching import Cache
import pytz
from dependency_injector.wiring import inject, Provide
from ..utils.hcaptcha import require_hcaptcha
from ..containers import Container
from ..services import TranslationService
from ..exceptions import (
    MissingKeyError,
    InvalidTypeError,
    EmptyTranslationError,
    SpellingSystemUnavailableError,
    SpellingSystemNotFoundError,
    TranslationNotFoundError
)


bp = Blueprint("translations", __name__, url_prefix='/api/v1')

@bp.route('/translations', methods=['POST'])
@require_hcaptcha
@inject
def translate(translation_service: TranslationService = Provide[Container.translation_service]):
    """
    Endpoint to translate traditional English spelling to alternative English spelling and vice
    versa.

    This endpoint accepts a POST request with a JSON request body and returns the translated
    English text in the JSON Response. Optionally, it saves the resulting translation in the
    database.

    JSON Request Parameters:
    - spellingSystem (str): Name of spelling system used for the translation.
    - forward (bool): Indicates the direction of translation.
    - text (str): Text to be translated.
    - save (bool): Indicates whether save the resulting translation.

    JSON Response Parameters:
    - id (uuid): ID of the translation. (only if 'save' was True in the request)
    - creation_date (DateTime): Date and time translation was inserted into the database. (only if
                                'save' was True in the request)
    - spellingSystem.name (str): Name of spelling system used for the translation.
    - spellingSystem.version (str): Version of spelling system used for the translation.
    - forward (bool): Indicates the direction of translation.
    - traditional_text (str): Text in traditional English spelling.
    - respelled_text (str): Text in alternative English spelling.

    Returns:
        Response: A JSON Response object containing the translated English text.

    Example:

        Request:
        POST /api/v1/translations
        Request Body: {
            "spellingSystem": "lytspel",
            "forward": True,
            "text": "Hello world!",
            "save": True
        }

        Response:
        POST /api/v1/translations
        Response Body: {
            "id": "7d9be066-6a0b-4459-9242-86dce2df6775",
            "creation_date": "2020-10-21T05:39:20+00:00",
            "spellingSystem": {
                "name": "lytspel",
                "version": "0.1.0"
            }
            "forward": True,
            "traditional_text": "Hello world!",
            "respelled_text": "Heló wurld!"
        }

    HTTP Status Codes:
    - 200 OK: Translated English text is returned.
    - 400 Bad Request: JSON request is malformed or requested spelling system is unavailable.
    """
    data = request.json

    save = data.get('save')
    spelling_system = data.get('spellingSystem')
    forward = data.get('forward')
    text = data.get('text')

    try:
        translation = translation_service.translate(
            spelling_system,
            forward,
            text,
            save
        )
    except (
        MissingKeyError,
        InvalidTypeError,
        EmptyTranslationError,
        NotImplementedError,
        SpellingSystemUnavailableError,
        SpellingSystemNotFoundError
    ) as e:
        return {'error': str(e)}, 400

    resp = {
        'spellingSystem': {
            'name': translation.spelling_system.name,
            'version': translation.spelling_system.version
        },
        'forward': translation.forward,
        'traditionalText': translation.traditional_text,
        'respelledText': translation.respelled_text
    }

    if save:
        resp['id'] = translation.id
        resp['creationDate'] = pytz.utc.localize(translation.creation_date).isoformat()

    return resp

@bp.route('/translations/<uuid:translation_id>', methods=['GET'])
@inject
def get_translation(
    translation_id: uuid,
    translation_service: TranslationService = Provide[Container.translation_service],
    cache: Cache = Provide[Container.cache]
):
    """
    Endpoint to get saved translation.

    This endpoint accepts a GET request with the appended translation ID (uuid).

    JSON Response Parameters:
    - id (uuid): ID of the translation.
    - creationDate (DateTime): Date and time translation was inserted into the database.
    - spellingSystem.name (str): Name of spelling system used for the translation.
    - spellingSystem.version (str): Version of spelling system used for the translation.
    - forward (bool): If true, traditional_text -> respelled_text. If false, respelledText -> \
        traditionalText.
    - traditionalText (str): Text in traditional English spelling (necessary if forward is True).
    - respelledText (str): Text in alternative English spelling (necessary if is_forward is False).

    Returns:
        Response: A JSON Response object containing the translated English text.

    Example:

        Request:
        GET /api/v1/translations/7d9be066-6a0b-4459-9242-86dce2df6775

        Response:
        GET /api/v1/translations
        Response Body: {
            "id": "7d9be066-6a0b-4459-9242-86dce2df6775",
            "creationDate": "2020-10-21T05:39:20-0700",
            "spellingSystem": {
                "name": "lytspel",
                "version": "0.2.0"
            }
            "forward": True,
            "traditionalText": "Hello world!",
            "respelledText": "Heló wurld!"
        }

    HTTP Status Codes:
    - 200 OK: Translated English text is returned.
    - 400 Bad Request: Translation ID is not a UUID.
    - 404 Not Found: Translation not found.
    """

    translation = cache.get(translation_id)

    if translation is None:
        try:
            translation = translation_service.get_translation_by_id(translation_id)
        except TranslationNotFoundError as e:
            return {'error': str(e)}, 404

        cache.set(translation_id, translation)

    resp = {
        'id': translation.id,
        'creationDate': pytz.utc.localize(translation.creation_date).isoformat(),
        'spellingSystem': {
            'name': translation.spelling_system.name,
            'version': translation.spelling_system.version
        },
        'forward': translation.forward,
        'traditionalText': translation.traditional_text,
        'respelledText': translation.respelled_text
    }

    return resp



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from dependency_injector import containers, providers
from flask_caching import Cache
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from .services import SpellingSystemService, TranslationService
from .repositories import SpellingSystemRepository, TranslationRepository


class Container(containers.DeclarativeContainer):  # pylint: disable=too-few-public-methods
    """Container for injecting dependencies into blueprint modules."""

    wiring_config = containers.WiringConfiguration(
        modules=[
            ".blueprints.spelling_system", 
            ".blueprints.translation",
            ".utils.populate_spelling_system_table"
        ]
    )

    db = providers.Singleton(SQLAlchemy)

    cache = providers.Singleton(Cache)

    migrate = providers.Singleton(Migrate)

    spelling_system_repository = providers.Singleton(
        SpellingSystemRepository,
        db=db
    )

    spelling_system_service = providers.Factory(
        SpellingSystemService,
        spelling_system_repository=spelling_system_repository
    )

    translation_repository = providers.Singleton(
        TranslationRepository,
        db=db
    )

    translation_service = providers.Factory(
        TranslationService,
        translation_repository=translation_repository,
        spelling_system_repository=spelling_system_repository
    )



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2025  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import uuid
from typing import List, Type
from flask import current_app
from .repositories import TranslationRepository, SpellingSystemRepository
from .model import SpellingSystem, Translation
from .exceptions import (
    MissingKeyError, InvalidTypeError, EmptyTranslationError, SpellingSystemUnavailableError
)


class SpellingSystemService:  # pylint: disable=too-few-public-methods
    """A service providing functionality for spelling system endpoints."""

    def __init__(
        self,
        spelling_system_repository: SpellingSystemRepository
    ):
        self._spelling_system_repository: SpellingSystemRepository = spelling_system_repository

    @staticmethod
    def get_enabled_spelling_systems() -> List[str]:
        """
        Returns:
            List: A List of enabled spelling systems.
        """
        return list(current_app.plugin_instances.keys())

    def get_enabled_spelling_system(self, name: str) -> SpellingSystem:
        """
        Returns:
            SpellingSystem: A SpellingSystem object representing the queried database record.
        """

        plugin_instance = current_app.plugin_instances.get(name)

        if plugin_instance is None:
            raise SpellingSystemUnavailableError(name)

        name = current_app.plugin_instances.get(name).name
        version = current_app.plugin_instances.get(name).version

        return self._spelling_system_repository.get(name, version)

    def get_spelling_system(self, name: str, version: str) -> SpellingSystem:
        """
        Returns:
            SpellingSystem: A SpellingSystem object representing the queried database record.
        """

        # raises SpellingSystemNotFoundError if not found
        return self._spelling_system_repository.get(name, version)

    def add_spelling_system(
        self,
        name: str,
        version: str,
        pretty_name: str,
        facts: str | None
    ) -> SpellingSystem:
        """
        Add a spelling system to the database.

        Args:
            name (str): Name of the spelling system to add.
            pretty_name (str): Pretty name of the spelling system to add.
            facts (str): JSON string for miscellaneous structured info about the spelling system
                plugin.

        Returns:
            SpellingSystem: A SpellingSystem object representing the added database record.
        """
        return self._spelling_system_repository.add(name, version, pretty_name, facts)

class TranslationService:
    """A service providing functionality for translation endpoints."""

    def __init__(
        self,
        translation_repository: TranslationRepository,
        spelling_system_repository: SpellingSystemRepository
    ):
        self._translation_repository: TranslationRepository = translation_repository
        self._spelling_system_repository: SpellingSystemRepository = spelling_system_repository

    def get_translation_by_id(self, translation_id: uuid) -> Translation:
        """
        Retrieve a Translation by id, from the database.

        Args:
            translation_id (uuid): Id of the Translation to query.

        Returns:
            Translation: A Translation object representing the queried database record.
        """
        return self._translation_repository.get_by_id(translation_id)

    def translate(
        self,
        name: str,
        forward: bool,
        text: str,
        save: bool
    ) -> Translation:
        """
        Perform a translation, optionally saving it to the database.

        Args:
            name (str): Name of the spelling system to use for translation.
            forward (bool): True for translation to the alternative spelling system. False for \
                translation to traditional English spelling.
            text (str): Text to be translated.
            save (bool): If true, persist the translation to the database.

        Returns:
            Translation: A Translation object representing the added database record.
        """

        # assign default save value
        if save is None:
            save = False

        def validate_key(key, key_pascal_case: str, cls: Type):
            if key is None:
                raise MissingKeyError(key_pascal_case)

            if not isinstance(key, cls):
                raise InvalidTypeError(key_pascal_case, cls)

        # exception handling
        validate_key(name, "spellingSystem", str)
        validate_key(forward, "forward", bool)
        validate_key(text, "text", str)

        if text == '':
            raise EmptyTranslationError

        plugin_instance = current_app.plugin_instances.get(name)

        if plugin_instance is None:
            raise SpellingSystemUnavailableError(name)

        spelling_system = self._spelling_system_repository.get(
            plugin_instance.name,
            plugin_instance.version
        )

        # get translation functions
        translate_to_respelling = plugin_instance.translate_to_respelling
        translate_to_traditional_spelling = plugin_instance.translate_to_traditional_spelling

        translation_length_limit = current_app.config['TRANSLATION_LENGTH_LIMIT']

        text = text[:translation_length_limit]

        if forward:
            traditional_text = text

            # raises NotImplementedFwdError if unimplemented
            respelled_text = translate_to_respelling(text)
        else:
            # raises NotImplementedBwdError if unimplemented
            traditional_text = translate_to_traditional_spelling(text)

            respelled_text = text

        translation = Translation(
            forward=forward,
            traditional_text=traditional_text,
            respelled_text=respelled_text,
            spelling_system_id=spelling_system.id
        )

        translation.spelling_system = spelling_system

        if save:
            translation = self._translation_repository.add(
                forward=forward,
                traditional_text=traditional_text,
                respelled_text=respelled_text,
                spelling_system_id=spelling_system.id,
            )

        return translation



'''
    Altspell  Flask web app for translating traditional English to respelled
    English and vice versa
    Copyright (C) 2024  Nicholas Johnson

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from flask import Blueprint, jsonify
from dependency_injector.wiring import Provide, inject
from ..services import SpellingSystemService
from ..containers import Container
from ..exceptions import SpellingSystemNotFoundError, SpellingSystemUnavailableError


bp = Blueprint("spelling_systems", __name__, url_prefix='/api/v1')

@bp.route('/spelling-systems', methods=['GET'])
@inject
def get_enabled_spelling_systems(spelling_system_service: SpellingSystemService = \
    Provide[Container.spelling_system_service]):
    """
    Endpoint that returns a list of enabled spelling systems.

    This endpoint accepts a GET request and returns a list of enabled spelling systems in the \
        JSON Response.

    Returns:
        Response: A JSON Response object containing a list of enabled spelling systems.

    Example:

        Request:
        GET /api/v1/spelling-systems

        Response:
        GET /api/v1/spelling-systems
        Response Body: [
            "lytspel",
            "soundspel"
        ]

    HTTP Status Codes:
    - 200 OK: List of spelling systems is returned.
    - 404 Not Found: Spelling system plugin not found.
    """
    try:
        spelling_systems = spelling_system_service.get_enabled_spelling_systems()
    except SpellingSystemUnavailableError as e:
        return {'error': str(e)}, 404

    return jsonify(spelling_systems)

@bp.route('/spelling-systems/<string:name>', methods=['GET'])
@inject
def get_enabled_spelling_system(
    name: str,
    spelling_system_service: SpellingSystemService = Provide[Container.spelling_system_service]
):
    """
    Endpoint that returns a JSON representation of a spelling system plugin.

    This endpoint accepts a GET request and returns a JSON representation of a spelling system \
        plugin.

    Returns:
        Response: A JSON response object containing a spelling system plugin.

    Example:

        Request:
        GET /api/v1/spelling-systems/lytspel

        Response:
        GET /api/v1/spelling-systems/lytspel
        Response Body: {
            "name": "lytspel",
            "prettyName": "Lytspel",
            "version": "0.2.1",
            "facts": {
                "pluginAuthor": "Nicholas Johnson",
                "spellingSystemAuthor": "Christian Siefkes"
            }
        }

    HTTP Status Codes:
    - 200 OK: Spelling system plugin is returned.
    - 404 Not Found: Spelling system plugin not found.
    """
    try:
        spelling_system = spelling_system_service.get_enabled_spelling_system(name)
    except SpellingSystemUnavailableError as e:
        return {'error': str(e)}, 404

    resp = {
        'name': spelling_system.name,
        'version': spelling_system.version,
        'prettyName': spelling_system.pretty_name,
    }

    if spelling_system.facts is not None:
        resp['facts'] = spelling_system.facts

    return resp

@bp.route('/spelling-systems/<string:name>/<string:version>', methods=['GET'])
@inject
def get_spelling_system(
    name: str,
    version: str,
    spelling_system_service: SpellingSystemService = Provide[Container.spelling_system_service]
):
    """
    Endpoint that returns a JSON representation of a spelling system plugin.

    This endpoint accepts a GET request and returns a JSON representation of a spelling system \
        plugin.

    Returns:
        Response: A JSON response object containing a spelling system plugin.

    Example:

        Request:
        GET /api/v1/spelling-systems/lytspel/0.1.0

        Response:
        GET /api/v1/spelling-systems/lytspel/0.1.0
        Response Body: {
            "name": "lytspel",
            "prettyName": "Lytspel",
            "version": "0.1.0",
            "facts": {
                "pluginAuthor": "Nicholas Johnson",
                "spellingSystemAuthor": "Christian Siefkes"
            }
        }

    HTTP Status Codes:
    - 200 OK: Spelling system plugin is returned.
    - 404 Not Found: Spelling system plugin not found.
    """
    try:
        spelling_system = spelling_system_service.get_spelling_system(name, version)
    except SpellingSystemNotFoundError as e:
        return {'error': str(e)}, 404

    resp = {
        'name': spelling_system.name,
        'version': spelling_system.version,
        'prettyName': spelling_system.pretty_name,
    }

    if spelling_system.facts is not None:
        resp['facts'] = spelling_system.facts

    return resp
